@@ -3,9 +3,14 @@ import logging
 import sys
 import uuid
 import argparse
+import readline
+import subprocess
+from bashutils.colors import color_text
 from mfcloud.attach import Terminal, AttachStdinProtocol
+from mfcloud.shell import BufferAwareCompleter
 from mfcloud.sync.client import FileServerError
 from mfcloud.sync.storage import get_storage, storage_sync
+from mfcloud.util import txtimeout
 
 import re
 import os
@@ -52,8 +57,7 @@ arg_parser = argparse.ArgumentParser(
     add_help=False
 )
 arg_parser.add_argument('-v', '--verbose', help='Show more logs', action='store_true', default=False)
-arg_parser.add_argument('-e', '--env', help='Environment to use', default='dev')
-arg_parser.add_argument('-h', '--host', help='Host to use', default='127.0.0.1')
+arg_parser.add_argument('-h', '--host', help='Host to use', default=None)
 arg_parser.add_argument(
     '-V', '--version',
     action='version',
@@ -61,22 +65,32 @@ arg_parser.add_argument(
 
 subparsers = arg_parser.add_subparsers()
 
+command_settings = {}
 
-def cli(help_, arguments=None):
+
+def cli(help_, arguments=None, by_ref=False, name=None):
     def cmd_decorator(func):
 
-        cmd = subparsers.add_parser(func.__name__, help=help_)
+        cmd = subparsers.add_parser(name or func.__name__, help=help_)
+
+        command_settings[func.__name__] = {
+            'need_app': False
+        }
+
+        if arguments:
+            command_settings[func.__name__]['need_app'] = arguments[0][0][0] == 'app'
 
         if arguments:
             for argument in arguments:
                 cmd.add_argument(*argument[0], **argument[1])
 
-        cmd.set_defaults(func=func.__name__)
+        if by_ref:
+            cmd.set_defaults(func=func)
+        else:
+            cmd.set_defaults(func=func.__name__)
 
         def _run(*args, **kwargs):
             func1 = func(*args, **kwargs)
-            if reactor.running:
-                reactor.stop()
             return func1
 
         return _run
@@ -89,8 +103,7 @@ def arg(*args, **kwargs):
 
 
 class ApiRpcClient(object):
-
-    def __init__(self, host='0.0.0.0', port=7080, settings=None):
+    def __init__(self, host='127.0.0.1', port=7080, settings=None):
         self.host = host
         self.port = port
         self.settings = settings
@@ -101,8 +114,9 @@ class ApiRpcClient(object):
 
         client = Client(host=self.host, settings=self.settings)
         try:
-            yield client.connect()
-
+            def _connect_failed():
+                raise Exception('Can\'t connect to the server on host %s' % self.host)
+            yield txtimeout(client.connect(), 3, _connect_failed)
             task = Task(task_name)
             task.on_progress = self.print_progress
 
@@ -152,7 +166,9 @@ class ApiRpcClient(object):
 
         client = Client(host=self.host, settings=self.settings)
         try:
-            yield client.connect()
+            def _connect_failed():
+                raise Exception('Can\'t connect to the server on host %s' % self.host)
+            yield txtimeout(client.connect(), 3, _connect_failed)
 
             task = Task(task_name)
             task.on_progress = self.print_progress
@@ -177,18 +193,101 @@ class ApiRpcClient(object):
         finally:
             stream_proto.stop()
 
-    @cli('Attach to container', arguments=(
-        arg('container_id', help='Container name'),
-    ))
-    def attach(self, container_id, **kwargs):
-        return self._exec_remote_with_pty('attach', container_id)
+
+    def print_app_details(self, app):
+
+        out = '\n'
+
+        x = PrettyTable(["Service name", "status", "ip", "cpu %", "memory", "volumes", "public urls"], hrules=ALL)
+
+        for service in app['services']:
+            if service['created']:
+                service_status = 'ON' if service['running'] else 'OFF'
+            else:
+                service_status = 'NOT CREATED'
+
+            volumes = ''
+            if service['volumes']:
+                volumes = '\n'.join(service['volumes'])
+
+            web = []
+
+            if app['status'] != 'error':
+                if 'web_service' in app and app['web_service'] == service['name']:
+                    web.append('http://' + app['fullname'])
+
+                    if 'public_urls' in app and app['public_urls']:
+                        for url in app['public_urls']:
+                            web.append('http://' + url)
+
+            x.add_row([
+                service['name'],
+                service_status,
+                service['ip'],
+                ("%.2f" % float(service['cpu'])) + '%',
+                str(service['memory']) + 'M',
+                volumes,
+                '\n'.join(web)
+            ])
+
+        out += str(x)
+
+        return out
+
+
+    def print_app_list(self, data):
+
+        x = PrettyTable(["Application name", "status", "cpu %", "memory", "Web", "Path"], hrules=ALL)
+        for app in data:
+
+            app_cpu = 0.0
+            app_mem = 0
+            web = ''
+
+            for service in app['services']:
+                app_mem += int(service['memory'])
+                app_cpu += float(service['cpu'])
+
+            if app['running']:
+                app_status = app['status']
+                services_cpu_list = ('%.2f' % app_cpu) + '%'
+                services_memory_list = str(app_mem) + 'M'
+
+            elif app['status'] == 'error':
+                app_status = 'ERROR: %s' % app['message']
+                web = app['message']
+                services_cpu_list = ''
+                services_memory_list = ''
+            else:
+                app_status = ''
+                services_cpu_list = ''
+                services_memory_list = ''
+
+            if app['status'] != 'error':
+                web_service_ = 'No web'
+                if 'web_service' in app and app['web_service']:
+                    web_service_ = app['web_service']
+                    if web_service_.endswith(app['name']):
+                        web_service_ = web_service_[0:-len(app['name']) - 1]
+
+                web = '%s -> [%s]' % (app['fullname'], web_service_)
+
+                if 'public_urls' in app and app['public_urls']:
+                    for url in app['public_urls']:
+                        web += '\n' + '%s -> [%s]' % (url, web_service_)
+
+            x.add_row([app['name'], app_status, services_cpu_list, services_memory_list, web, app['config']['path']])
+
+        return '\n' + str(x) + '\n'
 
     @cli('Run a command inside container', arguments=(
-        arg('name', help='Container name'),
+        arg('app', help='Application name', default=None, nargs='?'),
+        arg('service', help='Service name'),
         arg('command', help='Command to execute', default='/bin/bash', nargs='?'),
         arg('--no-tty', default=False, action='store_true', help='Disable tty binding'),
     ))
-    def run(self, name, command, no_tty=True, **kwargs):
+    def run(self, app, service, command, no_tty=True, **kwargs):
+        name = '%s.%s' % (service, require(app))
         return self._exec_remote_with_pty('run', name, command)
 
     @cli('Push appliction volume application', arguments=(
@@ -251,9 +350,6 @@ class ApiRpcClient(object):
         client.shutdown()
         yield sleep(0.01)
 
-        if self.stop_reactor:
-            reactor.stop()
-
     @cli('Creates a new application', arguments=(
         arg('name', help='App name'),
         arg('--user', default=None, help='Remote username'),
@@ -293,113 +389,54 @@ class ApiRpcClient(object):
         return self._remote_exec('init', name, os.path.realpath(path))
 
 
-    def print_app_details(self, app):
-
-        out = '\n'
-
-        x = PrettyTable(["Service name", "status", "ip",  "cpu %", "memory", "volumes", "public urls"], hrules=ALL)
-
-        for service in app['services']:
-            if service['created']:
-                service_status = 'ON' if service['running'] else 'OFF'
-            else:
-                service_status = 'NOT CREATED'
-
-            volumes = ''
-            if service['volumes']:
-                volumes = '\n'.join(service['volumes'])
-
-            web = []
-
-            if app['status'] != 'error':
-                if 'web_service' in app and app['web_service'] == service['name']:
-                    web.append('http://' + app['fullname'])
-
-                    if 'public_urls' in app and app['public_urls']:
-                        for url in app['public_urls']:
-                            web.append('http://' + url)
-
-            x.add_row([
-                service['name'],
-                service_status,
-                service['ip'],
-                ("%.2f" % float(service['cpu'])) + '%',
-                str(service['memory']) + 'M',
-                volumes,
-                '\n'.join(web)
-            ])
-
-        out += str(x)
-
-
-
-        return out
-
-
-    def print_app_list(self, data):
-
-        x = PrettyTable(["Application name", "status", "cpu %", "memory", "Web", "Path"], hrules=ALL)
-        for app in data:
-
-            app_cpu = 0.0
-            app_mem = 0
-            web = ''
-
-            for service in app['services']:
-                app_mem += int(service['memory'])
-                app_cpu += float(service['cpu'])
-
-            if app['running']:
-                app_status = app['status']
-                services_cpu_list = ('%.2f' % app_cpu) + '%'
-                services_memory_list = str(app_mem) + 'M'
-
-            elif app['status'] == 'error':
-                app_status = 'ERROR: %s' % app['message']
-                web = app['message']
-                services_cpu_list = ''
-                services_memory_list = ''
-            else:
-                app_status = ''
-                services_cpu_list = ''
-                services_memory_list = ''
-
-            if app['status'] != 'error':
-                web_service_ = 'No web'
-                if 'web_service' in app and app['web_service']:
-                    web_service_ = app['web_service']
-                    if web_service_.endswith(app['name']):
-                        web_service_ = web_service_[0:-len(app['name']) - 1]
-
-                web = '%s -> [%s]' % (app['fullname'], web_service_)
-
-                if 'public_urls' in app and app['public_urls']:
-                    for url in app['public_urls']:
-                        web += '\n' + '%s -> [%s]' % (url, web_service_)
-
-            x.add_row([app['name'], app_status, services_cpu_list, services_memory_list, web, app['config']['path']])
-
-        return '\n' + str(x) + '\n'
-
     @cli('List registered applications', arguments=(
-        arg('name', default=None, help='Application name', nargs='?'),
         arg('-f', '--follow', default=False, action='store_true', help='Continuously run list command'),
     ))
     @inlineCallbacks
-    def list(self, name=None, follow=False, **kwargs):
+    def list(self, follow=False, **kwargs):
         self.last_lines = 0
 
         def _print(data):
 
             ret = 'App not found.'
 
-            if not name:
-                ret = self.print_app_list(data)
-            else:
-                for app in data:
-                    if app['name'] == name:
-                        ret = self.print_app_details(app)
-                        break
+            ret = self.print_app_list(data)
+
+            if self.last_lines > 0:
+                print '\033[1A' * self.last_lines
+
+            print ret
+
+            self.last_lines = ret.count('\n') + 2
+
+        if follow:
+            while follow:
+                ret = yield self._remote_exec('list')
+                _print(ret)
+                yield sleep(1)
+        else:
+            ret = yield self._remote_exec('list')
+            _print(ret)
+
+    @cli('Show application status', arguments=(
+        arg('app', help='Application name', default=None, nargs='?'),
+        arg('-f', '--follow', default=False, action='store_true', help='Continuously run list command'),
+    ))
+    @inlineCallbacks
+    def status(self, app=None, follow=False, **kwargs):
+
+        name = require(app)
+
+        self.last_lines = 0
+
+        def _print(data):
+
+            ret = 'App not found.'
+
+            for app in data:
+                if app['name'] == name:
+                    ret = self.print_app_details(app)
+                    break
 
             if self.last_lines > 0:
                 print '\033[1A' * self.last_lines
@@ -438,8 +475,8 @@ class ApiRpcClient(object):
         else:
 
             table = Texttable(max_width=120)
-            table.set_cols_dtype(['t',  'a'])
-            table.set_cols_width([20,  100])
+            table.set_cols_dtype(['t', 'a'])
+            table.set_cols_width([20, 100])
 
             rows = [["Name", "Value"]]
             for name, val in data.items():
@@ -455,7 +492,7 @@ class ApiRpcClient(object):
         data = yield self._remote_exec('dns')
 
         table = Texttable(max_width=120)
-        table.set_cols_dtype(['t',  'a'])
+        table.set_cols_dtype(['t', 'a'])
         #table.set_cols_width([20,  100])
 
         rows = [["Name", "Value"]]
@@ -481,20 +518,31 @@ class ApiRpcClient(object):
         data = yield self._remote_exec('destroy', name)
         print 'result: %s' % pprintpp.pformat(data)
 
+    def format_app_srv(self, app, service):
+        require(app)
+        if service:
+            name = '%s.%s' % (service, app)
+        else:
+            name = app
+        return name
+
     @cli('Start containers', arguments=(
-        arg('name', help='Container or app name'),
+        arg('app', help='Application name', default=None, nargs='?'),
+        arg('service', help='Service name', default=None, nargs='?'),
     ))
     @inlineCallbacks
-    def start(self, name, **kwargs):
-        data = yield self._remote_exec('start', name)
+    def start(self, app, service, **kwargs):
+        print app, service
+        data = yield self._remote_exec('start', self.format_app_srv(app, service))
         print 'result: %s' % pprintpp.pformat(data)
 
     @cli('Create containers', arguments=(
-        arg('name', help='Container or app name'),
+        arg('app', help='Application name', default=None, nargs='?'),
+        arg('service', help='Service name', default=None, nargs='?'),
     ))
     @inlineCallbacks
-    def create(self, name, **kwargs):
-        data = yield self._remote_exec('create', name)
+    def create(self, app, service, **kwargs):
+        data = yield self._remote_exec('create', self.format_app_srv(app, service))
         print 'result: %s' % pprintpp.pformat(data)
 
     def format_domain(self, domain, ssl):
@@ -503,12 +551,14 @@ class ApiRpcClient(object):
         return domain
 
     @cli('Publish an application', arguments=(
-        arg('domain', help='Domain to publish'),
         arg('app', help='Application name'),
+        arg('domain', help='Domain to publish'),
+
         arg('--ssl', default=False, action='store_true', help='Ssl protocol'),
     ))
     @inlineCallbacks
     def publish(self, domain, app, ssl=False, **kwargs):
+        require(app)
         data = yield self._remote_exec('publish', self.format_domain(domain, ssl), app)
         print 'result: %s' % pprintpp.pformat(data)
 
@@ -542,7 +592,7 @@ class ApiRpcClient(object):
         self.on_vars_result(data)
 
     @cli('Unpublish an application', arguments=(
-        arg('domain', help='Domain to publish'),
+        arg('domain', help='Domain to unpublish'),
         arg('--ssl', default=False, action='store_true', help='Ssl protocol'),
     ))
     @inlineCallbacks
@@ -551,19 +601,21 @@ class ApiRpcClient(object):
         print 'result: %s' % pprintpp.pformat(data)
 
     @cli('Start application', arguments=(
-        arg('name', help='App name'),
+        arg('app', help='Application name', default=None, nargs='?'),
+        arg('service', help='Service name', default=None, nargs='?'),
     ))
     @inlineCallbacks
-    def restart(self, name, **kwargs):
-        data = yield self._remote_exec('restart', name)
+    def restart(self, app, service, **kwargs):
+        data = yield self._remote_exec('restart', self.format_app_srv(app, service))
         print 'result: %s' % pprintpp.pformat(data)
 
     @cli('Rebuild application', arguments=(
-        arg('name', help='App name'),
+        arg('app', help='Application name', default=None, nargs='?'),
+        arg('service', help='Service name', default=None, nargs='?'),
     ))
     @inlineCallbacks
-    def rebuild(self, name, **kwargs):
-        data = yield self._remote_exec('rebuild', name)
+    def rebuild(self, app, service, **kwargs):
+        data = yield self._remote_exec('rebuild', self.format_app_srv(app, service))
         print 'result: %s' % pprintpp.pformat(data)
 
 
@@ -598,16 +650,23 @@ class ApiRpcClient(object):
         return d
 
     @cli('Stop application', arguments=(
-        arg('name', help='App name'),
+        arg('app', help='Application name', default=None, nargs='?'),
+        arg('service', help='Service name', default=None, nargs='?'),
     ))
     @inlineCallbacks
-    def stop(self, name, **kwargs):
-        data = yield self._remote_exec('stop', name)
+    def stop(self, app, service, **kwargs):
+        data = yield self._remote_exec('stop', self.format_app_srv(app, service))
         print 'result: %s' % pprintpp.pformat(data)
 
 
-def main(argv):
+def require(app):
+    if not app:
+        raise ValueError(
+            'Application name is required. Pass it as an argument, or set it with "use" command in shell mode.')
+    return app
 
+
+def main(argv):
     console_handler = logging.StreamHandler(stream=sys.stderr)
     console_handler.setFormatter(logging.Formatter())
     console_handler.setLevel(logging.DEBUG)
@@ -619,15 +678,6 @@ def main(argv):
 
     logging.getLogger("requests").propagate = False
 
-    args = arg_parser.parse_args()
-
-
-    if args.verbose:
-        log.startLogging(sys.stdout)
-
-    args.argv0 = argv[0]
-
-
     class SslConfiguration(Configuration):
         enabled = False
         key = '/etc/mfcloud/ssl.key'
@@ -647,41 +697,171 @@ def main(argv):
 
     settings = MyAppConfiguration.load()
 
-    client = ApiRpcClient(host=args.host, settings=settings)
+    # client = ApiRpcClient(host=args.host, settings=settings)
 
-    if isinstance(args.func, str):
-        log.msg('Starting task: %s' % args.func)
+    if len(argv) < 2:
+        # Register our completer function
+        readline.set_completer(BufferAwareCompleter(
+            {'list': ['files', 'directories'],
+             'print': ['byname', 'bysize'],
+             'stop': [],
+            }).complete)
 
-        def ok(result):
+        # Use the tab key for completion
+        readline.parse_and_bind('tab: complete')
+
+        state = {
+            'app': None,
+            'host': 'me',
+        }
+
+
+        def green(text):
+            print(color_text(text, color='green'))
+
+        def yellow(text):
+            print(color_text(text, color='blue', bcolor='yellow'))
+
+        def info(text):
+            print()
+
+        @cli('Set current application', by_ref=True, arguments=(
+            arg('name', help='Application name', default=None, nargs='?'),
+        ))
+        def use(name, **kwargs):
+            if '@' in name:
+                app, host = name.split('@')
+                if host.strip() == '':
+                    host = 'me'
+                if app.strip() == '':
+                    app = None
+
+                state['app'] = app
+                state['host'] = host
+            else:
+                state['app'] = name
+
+
+        @inlineCallbacks
+        def mfcloud_shell():
+
+            intro = '''
+                     _                 _       ,--.
+     _ __ ___   ___ | | ___  _   _  __| |    _(    )
+    | '_ ` _ \ / __|| |/ _ \| | | |/ _` |   (    _'-. _
+    | | | | | | (__ | | (_) | |_| | (_| |    (       ) ),--.
+    |_| |_| |_|\___||_|\___/ \__,_|\__,_|  (                )-._
+                                               _________________)
+     Developer freindly cloud.
+            '''
+            intro = '\n'.join([x.ljust(80) for x in intro.split('\n')])
+
+            histfile = os.path.join(os.path.expanduser("~"), ".mcloud_history")
+            try:
+                readline.read_history_file(histfile)
+            except IOError:
+                pass
+
+            print(color_text(intro, color='white', bcolor='blue'))
+
+            line = ''
+            while line != 'exit':
+                print('')
+                prompt = 'mcloud: %s@%s> ' % (state['app'] or '~', state['host'])
+
+                try:
+                    line = None
+                    line = raw_input(color_text(prompt, color='white', bcolor='blue') + ' ')
+
+                    if line == '':
+                        continue
+
+                    if line == 'exit':
+                        break
+                    params = line.split(' ')
+                    args = arg_parser.parse_args(params)
+
+                    args.argv0 = argv[0]
+
+                    if args.host:
+                        host = args.host
+                    elif state['host'] == 'me' or not state['host']:
+                        host = '127.0.0.1'
+                    else:
+                        host = state['host']
+                    client = ApiRpcClient(host=host, settings=settings)
+
+                    for key, val in state.items():
+                        if not hasattr(args, key) or not getattr(args, key):
+                            setattr(args, key, val)
+
+                    if isinstance(args.func, str):
+                        yield getattr(client, args.func)(**vars(args))
+                    else:
+                        yield args.func(**vars(args))
+
+                except SystemExit:
+                    pass
+                except KeyboardInterrupt:
+                    continue
+                except EOFError:
+                    print('')
+                    break
+
+                except Exception as e:
+                    print color_text('Error:', color='white', bcolor='red')
+                    print(e)
+
+            readline.write_history_file(histfile)
             reactor.callFromThread(reactor.stop)
 
-        def err(failure):
-            print 'Unhandled failure: %s' % failure
-            reactor.callFromThread(reactor.stop)
 
-        d = getattr(client, args.func)(**vars(args))
+        # def heartbeat():
+        #     print "heartbeat"
+        #     reactor.callLater(1.0, heartbeat)
+        #
+        # reactor.callLater(1.0, heartbeat)
+
+        mfcloud_shell()
 
-        d.addCallback(ok)
-        d.addErrback(err)
         reactor.run()
-
-
     else:
-        args.func(**vars(args))
 
 
+        args = arg_parser.parse_args()
+
+        if args.verbose:
+            log.startLogging(sys.stdout)
+
+        args.argv0 = argv[0]
+
+        if isinstance(args.func, str):
+
+            try:
+                log.msg('Starting task: %s' % args.func)
+
+                def ok(result):
+                    reactor.callFromThread(reactor.stop)
+
+                def err(failure):
+                    print color_text('Error:', color='white', bcolor='red')
+                    print failure.value
+                    print
+                    reactor.callFromThread(reactor.stop)
+
+                client = ApiRpcClient(host=args.host or '127.0.0.1', settings=settings)
+                d = getattr(client, args.func)(**vars(args))
+
+                d.addCallback(ok)
+                d.addErrback(err)
+                reactor.run()
+            except Exception as e:
+                print(e)
 
 
-# intro = '''
-#            __      _                 _
-# _ __ ___  / _| ___| | ___  _   _  __| |
-#| '_ ` _ \| |_ / __| |/ _ \| | | |/ _` |
-#| | | | | |  _| (__| | (_) | |_| | (_| |
-#|_| |_| |_|_|  \___|_|\___/ \__,_|\__,_|
-#
-#Cloud that loves your data.
-#
-#'''
+        else:
+            args.func(**vars(args))
+
 
 def entry_point():
     """Zero-argument entry point for use with setuptools/distribute."""
